/**
 * Window Options extension
 * v1.1 (v2 on e.g.o).
 * mathematical.coffee@gmail.com
 *
 * This extension adds 'minimize', 'maximize', ..., 'always on top'
 * etc options to the right-click menu of the title bar.
 *
 * It is meant to be compatible with the StatusTitleBar extension.
 *
 * FAQ:
 *   Q: How do I ensure this is loaded *after* Status Title Bar?
 *   A: Just turn it off and on a few times.
 *      (Interestingly if this is installed after Status Title Bar then it will
 *       always load after it).
 *
 * TODO:
 * - translations (how? surely system has these already??)
 * - send to next/previous workspace?
 */

const Gdk = imports.gi.Gdk;
const Lang = imports.lang;
const Meta = imports.gi.Meta;
const Mainloop = imports.mainloop;
const Wnck = imports.gi.Wnck;
const Signals  = imports.signals;
const Shell = imports.gi.Shell;

const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;


/* CONFIGURATION */

const config = {
    // if you want the more traditional
    // 'always on visible workspace'/'always on this workspace'
    // separate menu items or just one on/off toggle.
    workspace_toggle : true,
};

// Edit these if you want to change the text that appears
// (say you like 'Minimise' not 'Minimize', or 'Above' not 'Always on top').
const MINIMIZE                    = "Minimize";
const MAXIMIZE                    = "Maximize";
const RESTORE                     = "Restore";
const MOVE                        = "Move";
const RESIZE                      = "Resize";
const ALWAYS_ON_TOP               = "Always on top";
const ALWAYS_ON_VISIBLE_WORKSPACE = "Always on visible workspace";
const ALWAYS_ON_THIS_WORKSPACE    = "Always on this workspace";
const CLOSE_WINDOW                = "Close"; // closes the current *window* instead of whole app.

// This is what will be in the menu.
// If you don't want an item, change it to 'false'.
// If you want a separator, make sure the string starts with
// 'separator' and hasn't already occurred in toAdd.
let toAdd = {};
toAdd[MINIMIZE] = true;
toAdd[RESTORE] = true;
toAdd[MAXIMIZE] = true;
toAdd['separator'] = true;

toAdd[MOVE] = true;
toAdd[RESIZE] = true;
toAdd['separator1'] = true;

toAdd[ALWAYS_ON_TOP] = true;
toAdd[ALWAYS_ON_VISIBLE_WORKSPACE] = true;
// Don't change this line.
toAdd[ALWAYS_ON_THIS_WORKSPACE] = !config.workspace_toggle;
toAdd['separator2'] = true;  // separates the above and 'Quit'.
toAdd[CLOSE_WINDOW] = true;

/* if there are already at least LENGTH_CUTOFF items on the existing window
 * menu, then we'll make ours a submenu instead of a full one (e.g. when
 * we add to the epiphany menu in GNOME 3.2)
 */
const LENGTH_CUTOFF = 4;

/****************** CODE ********************/
function LOG() {
    let msg = arguments[0];
    if (arguments.length > 1) {
        [].shift.call(arguments);
        msg = ''.format.apply(msg, arguments);
    }
    //log(msg);
    return msg;
}

/* WindowMenu class.
 * feed it in a menu to patch (the actual menu, e.g. appMenu.menu)
 * If you want to add this to your own class, you need to override
 * _getCurrentWindowMutter and _getCurrentWindowWnck to return the Meta.Window
 *  or Wnck.Window that you want the actions to apply to.
 */
function WindowOptionsMenuPatch() {
	this._init.apply(this, arguments);
}

WindowOptionsMenuPatch.prototype = {
    _init: function (menu, toAdd, config) {
        this.menu = menu;
        this._items = toAdd;
        this._menuSignals = [];
        // TODO: params.parse(MINIMIZE: true, ...)
        this._config = config;

        if (menu) {
            this.init();
        }
    },

    init: function (menu) {
        this.menu = menu || this.menu;
        if (this.menu) {
            this._setupMenu();
            this._menuSignals.push(this.menu.connect('destroy', Lang.bind(this,
                this._onMenuDestroy)));
            return true;
        }
        LOG("WARNING: set this.menu before calling init().");
        return false;
    },

    /* when the menu we are patching is destroyed, we have to as well. */
    _onMenuDestroy: function () {
        this.destroy();
        this.emit('destroy');
    },

    destroy: function () {
        /* remove menu items */
        if (this.submenu) {
            //this.submenu.destroy(); // "Error in signals.js: No signal connection 1 found" ?
            this.submenu = null;
        }

        let i = this._menuSignals.length;
        if (this.menu) { // menu might have been destroyed.
            while (i--) {
                this.menu.disconnect(this._menuSignals[i]);
            }
            this._menuSignals = [];
            this.menu = null;
        }
    },

    _setupMenu: function () {
        LOG("populating menu. Current length: %d", this.menu.length);
        this.submenu = (this.menu.length >= LENGTH_CUTOFF ?
            new PopupMenu.PopupSubMenuMenuItem(_("Window Options")) : 
            new PopupMenu.PopupMenuSection());
        this.menu.addMenuItem(this.submenu, 0);
        /* hack since PopuPSubMenuMenuItem has a .menu but MenuSection doesn't */
        if (this.submenu.menu) { 
            this.submenu = this.submenu.menu; 
        }

        /* Add items */
        for (let buttonName in this._items) {
            if (this._items.hasOwnProperty(buttonName) && this._items[buttonName]) {
                if (buttonName.indexOf('separator') === 0) {
                    this._items[buttonName] = new PopupMenu.PopupSeparatorMenuItem();
                } else if (buttonName === ALWAYS_ON_TOP ||
                           (this._config.workspace_toggle && buttonName === ALWAYS_ON_VISIBLE_WORKSPACE)) {
                    // toggles
                    this._items[buttonName] = new PopupMenu.PopupSwitchMenuItem(buttonName, false);
                    this._items[buttonName].connect('toggled',
                        Lang.bind(this, this._onActivate, buttonName));
                } else {
                    this._items[buttonName] = new PopupMenu.PopupMenuItem(buttonName);
                    this._items[buttonName].connect('activate',
                        Lang.bind(this, this._onActivate, buttonName));
                }
                this.submenu.addMenuItem(this._items[buttonName]);
            }
        }

        /* Whenever menu is opened, update the toggles */
        /* Note: could listen to 'notify::above' but then I have to always track
         *  the current window & make sure it's connected, right?
         * In any case, I can't find a 'notify::on_all_workspaces' so that has
         * to use the menu-opened signal anyway - may as well do it for all.
         */
        if (this._items[ALWAYS_ON_TOP] ||
                (this._config.workspace_toggle && 
                 this._items[ALWAYS_ON_VISIBLE_WORKSPACE])) {
            this._menuSignals.push(this.menu.connect('open-state-changed',
                Lang.bind(this, this._updateToggles)));
        }
    },

    _getCurrentWindowMutter: function () {
        let windows = Shell.WindowTracker.get_default().focus_app.get_windows();
        for (let i = 0; i < windows.length; i++) {
            if (windows[i].has_focus()) {
                return windows[i];
            }
        }
        // didn't find it.
        return null;
    },

    _getCurrentWindowWnck: function () {
        return Wnck.Screen.get_default().get_active_window();
    },

    _onActivate: function (item, event, op) {
        let current_window = null;

        /* Ideally we want to use mutter for everything, but the JS
         * interface doesn't have always on top/always on visible workspace exposed,
         * so we have to use wnck until they do.
         */
        if (op === ALWAYS_ON_TOP                   ||
                op === ALWAYS_ON_VISIBLE_WORKSPACE ||
                op === ALWAYS_ON_THIS_WORKSPACE) {
            // go through wnck: not ideal, but these methods are not available through mutter yet.
            current_window = this._getCurrentWindowWnck();
        } else {
            current_window = this._getCurrentWindowMutter();
        }

        if (current_window) {
            /* The following are Mutter methods */
            if (op === MINIMIZE) {
                current_window.minimize();
            } else if (op === MAXIMIZE) {
                // NOTE: see Meta-1.0.gir for exported flags etc.
                current_window.maximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
            } else if (op === RESTORE) {
                current_window.unmaximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
            } else if (op === MOVE) {
                //current_window.keyboard_move();
                /* Use global.display.begin_grab_op for move and resize.
                 * (In GNOME 3.4+ see panel.js for details).
                 * It only works with Meta.later_add, because at time of clicking
                 * this dropdown menu has the current grab focus.
                 * Move the cursor to middle of the window.
                 */
                Mainloop.idle_add(function () {
                    let pointer = Gdk.Display.get_default().get_device_manager().get_client_pointer(),
                        [scr,,] = pointer.get_position(),
                        rect    = current_window.get_input_rect(),
                        x       = rect.x + rect.width/2,
                        y       = rect.y + rect.height/2;
                    pointer.warp(scr, x, y);
                    global.display.begin_grab_op(global.screen, current_window,
                        Meta.GrabOp.MOVING, false, true, 1, 0, global.get_current_time(),
                        x, y);
                    return false;
                });
            } else if (op === RESIZE) {
                //current_window.keyboard_size();
                Mainloop.idle_add(function () {
                    let pointer = Gdk.Display.get_default().get_device_manager().get_client_pointer(),
                        [scr,,] = pointer.get_position(),
                        rect    = current_window.get_input_rect(),
                        x       = rect.x + rect.width,
                        y       = rect.y + rect.height;
                    pointer.warp(scr, x, y);
                    global.display.begin_grab_op(global.screen, current_window,
                        Meta.GrabOp.RESIZING_SE, false, true, 1, 0, global.get_current_time(),
                        x, y);
                    return false;
                });
            } else if (op === CLOSE_WINDOW) {
                current_window.delete(global.get_current_time());

            /* The following are WNCK methods */
            } else if (op === ALWAYS_ON_TOP) {
                if (current_window.is_above()) {
                    current_window.unmake_above();
                } else {
                    current_window.make_above();
                }
            /* Note: there is both a `pin` and a `stick` method.
             * I don't notice any difference between the two, but according to
             * http://developer.gnome.org/libwnck/stable/WnckWindow.html#wnck-window-is-pinned
             * `pinned` : "whether window is on all workspace. Pinned state may change anytime a
             *   'workspace-changed' signal gets emitted, but not when 'state-changed' gets emitted
             * `sticky` : "whether a window is sticky. May change any time a "state-changed" signal
             *   gets emitted.
             *
             * Sticky means "stuck to the glass", i.e. does not scroll with viewport.
             * In GDK/GTK+, sticky means "stuck to the glass" and *also* that the window is on all
             * workspaces. But here it only means the viewport aspect of it.
             *
             * Hence I think 'on all workspaces' is 'pin'. (But this can change with workspace-changed?!)
             */
            } else if (op === ALWAYS_ON_VISIBLE_WORKSPACE) {
                if (this._config.workspace_toggle && current_window.is_pinned()) {
                    current_window.unpin();
                } else {
                    current_window.pin();
                }
            } else if (op === ALWAYS_ON_THIS_WORKSPACE) {
                current_window.unpin();
            }
        }
    }, // _onActivate

    _updateToggles: function (obj, open) {
        /* Update the 'above'/'on all workspaces' toggles in case the user has
         * set this not using the menus.
         */
        if (!open) {
            return;
        }

        let current_window = this._getCurrentWindowMutter();
        if (this._items[ALWAYS_ON_TOP] &&
                (this._items[ALWAYS_ON_TOP].state !== current_window.above)) {
            this._items[ALWAYS_ON_TOP].setToggleState(current_window.above);
        }

        if (this._items[ALWAYS_ON_VISIBLE_WORKSPACE] &&
                (this._items[ALWAYS_ON_VISIBLE_WORKSPACE].state !== current_window.is_on_all_workspaces())) {
            this._items[ALWAYS_ON_VISIBLE_WORKSPACE].setToggleState(current_window.is_on_all_workspaces());
        }
    }
}; // end WindowOptionsMenuPatch.prototype
Signals.addSignalMethods(WindowOptionsMenuPatch.prototype);

/* WindowOptionsMenuHandler class.
 * This tries to find the system title bar object to patch the WindowOptionsMenu
 * to. It also tries to listen for it being destroyed to re-add.
 */
function WindowOptionsMenuHandler() {
	this._init.apply(this, arguments);
}
WindowOptionsMenuHandler.prototype = {
    _init: function (toAdd, config) {
        this.menuPatch = null;
        this.menuHolder = null;
        this._startupID = null;
        this._destroyID = null;

        /* Try to get the 'Title Bar' object.
         * This is identified as the object with a ._targetApp property.
         */
        let children = Main.panel._leftBox.get_children(),
            i = children.length;
        while (i--) {
            if (children[i]._delegate &&
                    children[i]._delegate._targetApp !== undefined) {
                 // if _targetApp is set but NULL it will still pass this test.
                this.setMenu(children[i]._delegate);
                break;
            }
        }
        if (!this.menuHolder) {
            LOG("Couldn't find an appMenu-like object - what to do now?!")
        }
    },

    /* repatches item 'menuHolder' which has member .menu */
    setMenu: function(menuHolder) {
        this.menuHolder = menuHolder;

        /* destroy the current patch */
        if (this.menuPatch) {
            if (this._destroyID) {
                this.menuPatch.disconnect(this._destroyID);
            }
            this.menuPatch.destroy();
            this.menuPatch = null;
        }
        // UPTO: where to put the 'new '?
        this.menuPatch = new WindowOptionsMenuPatch(menuHolder.menu, toAdd, config);
        /* theMenu.menu is null until there is a focus app, so wait until then
         * to intialise the menu. Then we can stop listening to this event.
         */
        if (!menuHolder.menu) {
            let tracker = Shell.WindowTracker.get_default();
            this._startupID = tracker.connect('notify::focus-app', Lang.bind(this, function () {
                if (this.menuHolder.menu) {
                    this.menuPatch.init(this.menuHolder.menu);
                    Shell.WindowTracker.get_default().disconnect(this._startupID);
                    this._startupID = null;
                }
            }));
        }
        this._destroyID = this.menuPatch.connect('destroy', Lang.bind(this, this._onMenuDestroy));
    },

    _onMenuDestroy: function () {
        /* Epiphany destroys the menu & replace it with its own (GNOME 3.3+).
         * We wait for it to be made and re-patch it.
         * But *NOTE* you get a *really long* menu, very inconvenient.
         * Add it as a submenu 'Window' ? Or two-columns ?
         */
        Mainloop.idle_add(Lang.bind(this, function () {
            this.setMenu(this.menuHolder);
            return false;
        }));
    },
    
    destroy: function () {
        if (this._startupID) {
            let tracker = Shell.WindowTracker.get_default();
            tracker.disconnect(this._startupID);
            this._startupID = null;
        }
        if (this._destroyID) {
            this.menuPatch.disconnect(this._destroyID);
            this._destroyID = null;
        }
        this.menuPatch.destroy();
    }
}; // end WindowOptionsMenuHandler.prototype

let windowOptions = null;

function init() {
}

function enable() {
    windowOptions = new WindowOptionsMenuHandler();
}

function disable() {
    if (windowOptions) {
        windowOptions.destroy();
    }
    windowOptions = null;
}

